from io import FileIO, StringIO
from pathlib import Path
from .token import Token, TK
from .iterator import Iterator


def is_alpha(c: str) -> bool:
    if c:
        return 'a' <= c <= 'z' or 'A' <= c <= 'Z' or c == '_'
    return False

def is_numeric(c: str) -> bool:
    if c:
        return '0' <= c <= '9'
    return False

def is_alpha_numeric(c: str) -> bool:
    return is_alpha(c) or is_numeric(c)


class Lexer(Iterator):
    class EmptyStreamException(Exception):
        def __init__(self, *args):
            super().__init__(*args)

    def __init__(self, filename: str = None, filepath: Path = None, file: FileIO = None, stream: StringIO = None):

        if filename:
            filepath = Path(filename)

        if filepath:
            file = filepath.open('r')
        
        if file:
            stream = file.read()
        
        if not stream:
            raise Lexer.EmptyStreamException('No input given to lexer!')

        Iterator.__init__(self, iterable=stream)

        self.tokens : list[Token] = []

        # File position
        self.row : int = 0
        self.col : int = 0

        # Current token
        self.tbeg : int = 0
        self.tend : int = 0

    def next(self) -> str:
        c = super().next()

        self.col += 1

        if c == TK.LineFeed:
            self.row += 1
            self.col = 0

        return c

    def make_token(self, type: int) -> Token:
        self.tbeg = self.tend  # End of last token
        self.tend = self.pos + 1  # End of current token
        text = self.buffer[self.tbeg:self.tend] if self.tend - self.tbeg > 1 else self.buffer[self.pos]
        return Token(self.tbeg, self.tend, self.row, self.col, type, text)

    def tokenize(self) -> list[Token]:
        c : str = self.get()

        while self and c:

            if c in TK.Whitespaces:
                self.tokens.append(self.make_token(TK[c]))

            elif c in TK.Symbols:
                self.tokens.append(self.make_token(TK[c]))

            elif is_numeric(c):
                while is_numeric(c):
                    c = self.next()
                self.prev()

                self.tokens.append(self.make_token(TK.Number))

            elif is_alpha(c):
                while is_alpha_numeric(c):
                    c = self.next()
                self.prev()
                
                self.tokens.append(self.make_token(TK.Word))

            else:
                self.tokens.append(self.make_token(TK.Symbol))
            
            c = self.next()

        return self.tokens



from io import FileIO, StringIO
from pathlib import Path
from logging import Logger, getLogger
from struct import unpack

from parxel.token import Token, TK
from parxel.iterator import Iterator
from parxel.nodes import Node, Document
from parxel.lexer import Lexer


logger = getLogger(__name__)


class Parser(Iterator):
    def __init__(self,
        iterable: list = None,
        root: Node = None,
        filename: str = None,
        filepath: Path = None,
        file: FileIO = None,
        logger: Logger = logger):

        Iterator.__init__(self, iterable=iterable)

        self.logger = logger
        self.filename = filename
        self.file = file

        if filepath:
            self.filepath: Path = filepath
            self.root = Document(filepath)
        else:
            self.root = Node()
        
        if root:
            self.root = root

        # Current Node
        self.nbeg: int = 0
        self.nend: int = 0

    def parse(self) -> Node | Document:
        if self.filepath:
            self.logger.debug(f'Processing {self.filepath} ...')
        return self.parse_format()
    
    def parse_format(self):
        raise NotImplementedError('Implement the "parse_format" method!')
    
    @classmethod
    def read(cls: type, filename: str = None, filepath: Path = None, file: FileIO = None, stream: StringIO = None):

        lexer = Lexer(filename, filepath, file, stream)
        tokens: list[Token] = lexer.tokenize()

        obj = cls(filepath, tokens)
        return obj.parse()


class BinaryParser(Parser):
    def __init__(self,
        buffer: bytes = None,
        root: Node = None,
        filename: str = None,
        filepath: Path = None,
        file: FileIO = None,
        logger: Logger = logger):

        if filename:
            filepath = Path(filename)

        if filepath:
            file = filepath.open('rb')
        
        if file:
            buffer = file.read()

        Parser.__init__(self, iterable=buffer, root=root, filename=filename, filepath=filepath, file=file, logger=logger)

    def advance(self, distance: int) -> bytes:
        els : bytearray = bytearray(distance)
        for i in range(distance):
            els[i] = self.get() or 0
            self.next()
        return els

    def byte(self) -> int:
        return self.next() or 0

    def bytes(self, distance: int) -> bytes:
        return self.advance(distance)
    
    def int16(self, byteorder: str = 'little') -> int:
        return int.from_bytes(self.bytes(2), byteorder=byteorder)

    def int16_array(self, size: int, byteorder: str = 'little') -> list[int]:
        return [self.int16(byteorder) for _ in range(size)]
    
    def int32(self, byteorder: str = 'little') -> int:
        return int.from_bytes(self.bytes(4), byteorder=byteorder)

    def int32_array(self, size: int, byteorder: str = 'little') -> list[int]:
        return [self.int32(byteorder) for _ in range(size)]
    
    def int64(self, byteorder: str = 'little') -> int:
        return int.from_bytes(self.bytes(8), byteorder=byteorder)

    def int64_array(self, size: int, byteorder: str = 'little') -> list[int]:
        return [self.int64(byteorder) for _ in range(size)]
    
    def float32(self, byteorder: str = '<') -> float:
        return unpack(byteorder + 'f', self.bytes(4))[0]

    def float32_array(self, size: int, byteorder: str = '<') -> list[float]:
        return unpack(byteorder + ('f' * size), self.bytes(size * 4))
    
    def float64(self, byteorder: str = '<') -> float:
        return unpack(byteorder + 'd', self.bytes(8))[0]

    def float64_array(self, size: int, byteorder: str = '<') -> list[float]:
        return unpack(byteorder + ('d' * size), self.bytes(size * 8))
    
    def string(self, size: int, encoding: str = 'utf-8') -> str:
        return self.bytes(size).decode(encoding)
    
    def collect_bytes(self) -> bytes:
        self.nbeg = self.nend  # End of last node
        self.nend = self.pos  # End of current node
        return self.buffer[self.nbeg:self.nend]


class TextParser(Parser):
    class EmptyStreamException(Exception):
        def __init__(self, *args):
            super().__init__(*args)

    class UnexpectedTokenException(Exception):
        def __init__(self, *args):
            super().__init__(*args)

    def __init__(self,
        tokens: list[Token] = None,
        root: Node = None,
        filename: str = None,
        filepath: Path = None,
        file: FileIO = None,
        stream: StringIO = None,
        logger: Logger = logger):

        if filename:
            filepath = Path(filename)

        if filepath:
            file = filepath.open('r')
        
        if file:
            stream = file.read()
        
        if not stream:
            file_name = '' if file is None else f'"{file.name}"'
            logger.error(f'Empty stream {file_name}')
            raise TextParser.EmptyStreamException(f'No input given to parser! f{file_name}')

        if not tokens:
            lexer = Lexer(filename=filename, filepath=filepath, file=file, stream=stream)
            tokens = lexer.tokenize()

        Parser.__init__(self, iterable=tokens, root=root, filename=filename, filepath=filepath, file=file, logger=logger)

    def consume(self, type: TK) -> bool:
        if self.get() and self.get().type == type:
            self.next()
            return True
        return False

    def consume_strict(self, type: TK) -> None:
        if not self.consume(type):
            self.error(type)

    def consume_any(self, types: list[TK]) -> bool:
        if self.get() and self.get().type in types:
            self.next()
            return True
        return False

    def consume_until(self, type: TK) -> None:
        while self.get() and self.get().type != type:
            self.next()

    def consume_until_any(self, types: list[TK]) -> None:
        while self.get() and self.get().type not in types:
            self.next()

    def consume_while(self, type: TK) -> None:
        while self.get() and self.get().type == type:
            self.next()

    def consume_while_any(self, types: list[TK]) -> None:
        while self.get() and self.get().type in types:
            self.next()

    def discard(self) -> list[Token]:
        self.next()
        return self.collect_tokens()

    def number_of_tokens(self) -> int:
        return self.pos - self.nend

    def tokens(self) -> list[Token]:
        return self.buffer[self.nend:self.pos + 1]

    def collect_tokens(self) -> list[Token]:
        self.nbeg = self.nend  # End of last node
        self.nend = self.pos  # End of current node
        return self.buffer[self.nbeg:self.nend]

    def error(self, expected: TK) -> None:
        nl = '\n'
        t: Token = self.get()
        tokens = self.buffer[self.nend:self.pos + 1]
        text = "".join(list(map(lambda x: x.text, tokens)))
        indent = len(text[text.rfind(nl)]) if text.find('\n') > -1 else len(text) - len(tokens[-1].text)

        msg = f'\n\n{self.filepath.absolute()}: Line {t.row} Col {t.col}\n\n'
        msg += f'{text}\n'
        msg += f'{indent * " "}{"^" * len(self.get().text)}\n\n'
        msg += f'Expected \'{expected}\' got \'{tokens[-1].text}\'\n'
        msg += f'Last tokens: {self.tokens()}\n'

        self.logger.error(f'Unexpected token {self.filepath}: Expected \'{expected}\' got \'{tokens[-1].text}\'')
        raise TextParser.UnexpectedTokenException(msg)



class Iterator:
    def __init__(self, iterable: list):
        self.buffer : list = iterable
        self.pos : int = 0
        self.beg : int = 0
        self.end : int = len(self.buffer)

        if self.end > 0:
            self.el = self.buffer[0]

    def __bool__(self) -> bool:
        return self.pos < self.end

    def __getitem__(self, index: int):
        if not self or index >= self.end:
            return None

        return self.buffer[index]

    def get(self):
        if not self:
            return None

        self.el = self.buffer[self.pos]
        return self.el

    def next(self):
        self.pos += 1

        return self.get()
    
    def prev(self):
        if self.pos > 0:
            self.pos -= 1

        return self.get()

    def advance(self, distance: int) -> list:
        els : list = [None] * distance
        for i in range(distance):
            els[i] = self.get()
            self.next()
        return els

    def peek(self, distance: int = 1) -> list | None:
        if self.pos + distance < self.end:
            return self.buffer[self.pos + distance]
        return None
    
    def consume(self, el) -> bool:
        if self.buffer[self.pos] == el:
            self.next()
            return True
        return False

    def consume_any(self, el : list) -> bool:
        if self.buffer[self.pos] in el:
            self.next()
            return True
        return False

    def consume_until(self, el):
        while self and self.buffer[self.pos] != el:
            self.next()

    def consume_until_any(self, el : list):
        while self and self.buffer[self.pos] not in el:
            self.next()

    def consume_while(self, el):
        while self and self.buffer[self.pos] == el:
            self.next()

    def consume_while_any(self, el : list):
        while self and self.buffer[self.pos] in el:
            self.next()



from pathlib import Path
from hashlib import md5
from parxel.token import Token
from networkx import Graph


class Node:
    def __init__(self, parent=None):
        self.parent: Node = parent
        self.children: list[Node] = []
        self.scope: Node = self

        if parent:
            parent.add(self)

    def type(self) -> str:
        return self.__class__.__name__
    
    def hash(self, *tweak: str):
        hash_string = ''
        hash_string += self.type()
        hash_string += ''.join(str(arg) for arg in tweak if arg is not None)
        hash_string += str(len(self.children))
        hash_string += ''.join(set(map(lambda x : x.type(), self.children)))

        for child in self.children:
            hash_string += child.hash()
        
        return md5(hash_string.encode('utf-8')).hexdigest()

    def add(self, other) -> None:
        other.parent = self
        self.children.append(other)

    def enter_scope(self, other) -> None:
        self.scope.add(other)
        self.scope = other

    def exit_scope(self) -> None:
        self.scope = self.scope.parent

    def add_to_scope(self, other) -> None:
        self.scope.add(other)

    def dump(self, level: int = 0, properties: bool = False) -> str:
        s = f'{" " * level}{self.__class__.__name__:20s}\n'
        if properties:
            for k, v in self.__dict__.items():
                if k[0] != '_' and k[1] != '_':
                    s += f'{" " * level}- {k:20s} {v}\n'
        for c in self.children:
            s += c.dump(level + 1, properties)
        
        return s

    def walk(self):
        yield self

        for child in self.children:
            yield from child.walk()


class Folder(Node):
    def __init__(self, path: Path, parent: Node = None):
        Node.__init__(self, parent=parent)

        self.path: Path = path
    
    def hash(self, *tweak: str):
        return super().hash(str(self.path), tweak)


class Document(Node):
    def __init__(self, filepath: Path, parent: Node = None):
        Node.__init__(self, parent=parent)

        self.filepath: Path = filepath
    
    def hash(self, *tweak: str):
        return super().hash(str(self.filepath), tweak)


class LexicalNode(Node):
    def __init__(self, tokens: list[Token], parent: Node = None):
        Node.__init__(self, parent=parent)

        self.tokens: list[Token] = tokens
    
    def hash(self, *tweak: str):
        return super().hash(self.raw(), tweak)

    def raw(self) -> str:
        return ''.join(list(map(lambda x: x.text, self.tokens)))

    def dump(self, level: int = 0, properties: bool = False) -> str:
        s = f'{" " * level}{self.__class__.__name__:20s}{bytearray(self.raw(), encoding="utf-8")}\n'
        if properties:
            for k, v in self.__dict__.items():
                if k[0] != '_' and k[1] != '_':
                    s += f'{" " * level}- {k:20s} {v}\n'
        for c in self.children:
            s += c.dump(level + 1, properties)
        
        return s


class BinaryNode(Node):
    def __init__(self, blob: bytes, parent: Node = None):
        Node.__init__(self, parent=parent)

        self.bytes: bytes = blob
    
    def hash(self, *tweak: str):
        return super().hash(self.bytes, tweak)
